#ifndef COPAN_HPP_
#define COPAN_HPP_

#include "SolverTypes.h"

#ifdef COPAN

#include <iostream>
#include <string>
#include <stdint.h>
#include <fstream>
#include <map>

#define CoPAn_Do(X) X

namespace Minisat {

struct CoPAn{
    typedef std::map<Minisat::CRef,uint32_t>    cls2id_t;

    protected:
	uint32_t 								cls_id;
	std::ofstream*							cls_out;
	std::ofstream*							confl_out;

    
    cls2id_t                                cls2id;
	typedef cls2id_t::iterator              c2i_iter;
    typedef cls2id_t::const_iterator        cc2i_iter;

    //Helpers
    inline static int toDimacs(const Lit& l);
    inline uint32_t getId(const CRef& cref);
    
public:
    CoPAn():cls_id(0),cls_out(NULL),confl_out(NULL){}    
    
    //translate clause pointer to clause id for CoPAn
    inline static uint32_t getId(const cls2id_t& mp, const CRef& cref);
    
    //init the output files
    inline void setOutFiles(std::string bs);
	

    //register new clauses with some properties -> can be extended easily 
	inline void regCls(const CRef& cref,const vec<Lit>& lits, uint32_t lbd_dummy, uint32_t bj, float act);
	inline void regUnit(const Lit& flit, uint32_t bj);

    //produce the output for a conflict clause
    //add one reason to the current conflict
	inline void toConfl(const CRef& rsn); 
    //write current conflict out
	inline void finiConfl(){*confl_out << " : " <<cls_id << std::endl;}
	
    //remove clauses at reduceDB/ garbage collection
	inline void	gcStart(){*cls_out << "GC<" <<cls_id-1 <<"> ";}
	inline void	gcEnd(){*cls_out << std::endl;}
	inline void	gcHit(const CRef& del);//clause is deleted
    
    //change mapping of pointers to clause ids when minisat moves memory physically
    inline void movePtrs(std::map<Minisat::CRef,uint32_t>& old);
    inline void regCls(const CRef& cref, uint32_t cid);
};


extern CoPAn copan;

int CoPAn::toDimacs(const Lit& l){             
    int val =  var(l) + 1;
    return sign(l) ? -val : val;
}


void CoPAn::regCls(const CRef& cref, uint32_t cid){

    c2i_iter ex = cls2id.find(cref);
    
    if(ex != cls2id.end()){ 
        ex->second = cid; 
    }else{
        cls2id.insert(std::make_pair<CRef,uint32_t>(cref,cid));}    
}



void CoPAn::regCls(const CRef& cref,const vec<Lit>& lits, uint32_t lbd, uint32_t bj, float act){

	regCls(cref,cls_id);

    *cls_out << cls_id << ": ";
    for(int k=0;k!=lits.size();++k){
        *cls_out << toDimacs(lits[k])<<" ";}
    
    if(lbd != 0){   *cls_out << "l " << lbd << " ";}
    if(bj != 0){    *cls_out << "b " << bj << " ";}
    if(act != 0){   *cls_out << "a " << act << " ";}
   
   *cls_out        << std::endl;
    
    ++cls_id;
}




void CoPAn::regUnit(const Lit& flit, uint32_t bj){

	*cls_out << cls_id << ": ";
	*cls_out << toDimacs(flit) << " ";
	if(bj != 0){*cls_out << "b " << bj;}
    *cls_out << std::endl;

    ++cls_id;
}


uint32_t CoPAn::getId(const cls2id_t& mp, const CRef& cref){
    cc2i_iter ex = mp.find(cref);
    return ex != mp.end() ? ex->second : 0;
}
    

uint32_t CoPAn::getId(const CRef& cref){
	return getId(cls2id,cref);
}

void CoPAn::movePtrs(std::map<Minisat::CRef,uint32_t>& old){
    cls2id.swap(old);    
}
 
 
void CoPAn::gcHit(const CRef& del){
    *cls_out << getId(del) << " "; 
    cls2id.erase(del);
}
    


void CoPAn::setOutFiles(std::string bs){
 
	uint32_t p 		= bs.rfind(".cnf");
	if(p<bs.size()){
		bs			= bs.erase(p,bs.size());}

	std::string opn = bs + "_copan.clses";
	cls_out 		= new std::ofstream(opn.c_str());

	opn			    = bs + "_copan.confl" ;
	confl_out 		= new  std::ofstream(opn.c_str());

    printf("c Log Conflicts for CoPAn to %s", opn.c_str());
}


void CoPAn::toConfl(const CRef& rsn){
	uint32_t cid = getId(rsn);
	*confl_out << cid << " ";
}


}

#else
#define CoPAn_Do(X)
#endif


#endif 
